from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from exchange.checkpoint import CheckpointData
from exchange.message import Message
from exchange.moderators import PassiveModerator
from exchange.moderators.base import Moderator

if TYPE_CHECKING:
    from exchange.exchange import Exchange

# currently this is the point at which we start to truncate, so
# so once we get to this token size the token count will exceed this
# by a little bit.
# TODO: make this configurable for each provider
MAX_TOKENS = 100000


class ContextTruncate(Moderator):
    def __init__(
        self,
        model: Optional[str] = None,
        max_tokens: int = MAX_TOKENS,
    ) -> None:
        self.model = model
        self.system_prompt_token_count = 0
        self.max_tokens = max_tokens
        self.last_system_prompt = None

    def rewrite(self, exchange: Exchange) -> None:
        """Truncate the exchange messages with a FIFO strategy."""
        self._update_system_prompt_token_count(exchange)

        if exchange.checkpoint_data.total_token_count < self.max_tokens:
            return

        messages_to_remove = self._get_messages_to_remove(exchange)
        for _ in range(len(messages_to_remove)):
            exchange.pop_first_message()

    def _update_system_prompt_token_count(self, exchange: Exchange) -> None:
        is_different_system_prompt = False
        if self.last_system_prompt != exchange.system:
            is_different_system_prompt = True
            self.last_system_prompt = exchange.system

        if not self.system_prompt_token_count or is_different_system_prompt:
            # calculate the system prompt tokens (includes functions etc...)
            # we use a placeholder message with one token, which we subtract later
            # this ensures compatibility with providers that require a user message
            _system_token_exchange = exchange.replace(
                messages=[Message.user("a")],
                checkpoint_data=CheckpointData(),
                moderator=PassiveModerator(),
                model=self.model if self.model else exchange.model,
            )
            _system_token_exchange.generate()
            last_system_prompt_token_count = self.system_prompt_token_count
            self.system_prompt_token_count = _system_token_exchange.checkpoint_data.total_token_count - 1

            exchange.checkpoint_data.total_token_count -= last_system_prompt_token_count
            exchange.checkpoint_data.total_token_count += self.system_prompt_token_count

    def _get_messages_to_remove(self, exchange: Exchange) -> List[Message]:
        # this keeps all the messages/checkpoints
        throwaway_exchange = exchange.replace(
            moderator=PassiveModerator(),
        )

        # get the messages that we want to remove
        messages_to_remove = []
        while throwaway_exchange.checkpoint_data.total_token_count > self.max_tokens:
            _, messages = throwaway_exchange.pop_first_checkpoint()
            messages_to_remove.extend(messages)

        while len(throwaway_exchange.messages) > 0 and throwaway_exchange.messages[0].tool_result:
            # we would need a corresponding tool use once we resume, so we pop this one off too
            # and summarize it as well
            _, messages = throwaway_exchange.pop_first_checkpoint()
            messages_to_remove.extend(messages)
        return messages_to_remove



import json
import traceback
from copy import deepcopy
from typing import Any, Dict, List, Mapping, Tuple

from attrs import define, evolve, field, Factory
from tiktoken import get_encoding

from exchange.checkpoint import Checkpoint, CheckpointData
from exchange.content import Text, ToolResult, ToolUse
from exchange.message import Message
from exchange.moderators import Moderator
from exchange.moderators.truncate import ContextTruncate
from exchange.providers import Provider, Usage
from exchange.tool import Tool
from exchange.token_usage_collector import _token_usage_collector


def validate_tool_output(output: str) -> None:
    """Validate tool output for the given model"""
    max_output_chars = 2**20
    max_output_tokens = 16000
    encoder = get_encoding("cl100k_base")
    if len(output) > max_output_chars or len(encoder.encode(output)) > max_output_tokens:
        raise ValueError("This tool call created an output that was too long to handle!")


@define(frozen=True)
class Exchange:
    """An exchange of messages with an LLM

    The exchange class is meant to be largely immutable, with only the message list
    growing once constructed. Use .replace to alter the model, tools, etc.

    The exchange supports tool usage, calling tools and letting the model respond when
    using the .reply method. It handles most forms of errors and sends those errors back
    to the model, to let it attempt to recover.
    """

    provider: Provider
    model: str
    system: str
    moderator: Moderator = field(default=ContextTruncate())
    tools: Tuple[Tool] = field(factory=tuple, converter=tuple)
    messages: List[Message] = field(factory=list)
    checkpoint_data: CheckpointData = field(factory=CheckpointData)
    generation_args: dict = field(default=Factory(dict))

    @property
    def _toolmap(self) -> Mapping[str, Tool]:
        return {tool.name: tool for tool in self.tools}

    def replace(self, **kwargs: Dict[str, Any]) -> "Exchange":
        """Make a copy of the exchange, replacing any passed arguments"""
        # TODO: ensure that the checkpoint data is updated correctly. aka,
        # if we replace the messages, we need to update the checkpoint data
        # if we change the model, we need to update the checkpoint data (?)

        if kwargs.get("messages") is None:
            kwargs["messages"] = deepcopy(self.messages)
        if kwargs.get("checkpoint_data") is None:
            kwargs["checkpoint_data"] = deepcopy(
                self.checkpoint_data,
            )
        return evolve(self, **kwargs)

    def add(self, message: Message) -> None:
        """Add a message to the history."""
        if self.messages and message.role == self.messages[-1].role:
            raise ValueError("Messages in the exchange must alternate between user and assistant")
        self.messages.append(message)

    def generate(self) -> Message:
        """Generate the next message."""
        self.moderator.rewrite(self)
        message, usage = self.provider.complete(
            self.model,
            self.system,
            messages=self.messages,
            tools=self.tools,
            **self.generation_args,
        )
        self.add(message)
        self.add_checkpoints_from_usage(usage)  # this has to come after adding the response

        # TODO: also call `rewrite` here, as this will make our
        # messages *consistently* below the token limit. this currently
        # is not the case because we could append a large message after calling
        # `rewrite` above.
        # self.moderator.rewrite(self)

        _token_usage_collector.collect(self.model, usage)
        return message

    def reply(self, max_tool_use: int = 128) -> Message:
        """Get the reply from the underlying model.

        This will process any requests for tool calls, calling them immediately, and
        storing the intermediate tool messages in the queue. It will return after the
        first response that does not request a tool use

        Args:
            max_tool_use: The maximum number of tool calls to make before returning. Defaults to 128.
        """
        if max_tool_use <= 0:
            raise ValueError("max_tool_use must be greater than 0")
        response = self.generate()
        curr_iter = 1  # generate() already called once
        while response.tool_use:
            content = []
            for tool_use in response.tool_use:
                tool_result = self.call_function(tool_use)
                content.append(tool_result)
            self.add(Message(role="user", content=content))

            # We've reached the limit of tool calls - break out of the loop
            if curr_iter >= max_tool_use:
                # At this point, the most recent message is `Message(role='user', content=ToolResult(...))`
                response = Message.assistant(
                    f"We've stopped executing additional tool cause because we reached the limit of {max_tool_use}",
                )
                self.add(response)
                break
            else:
                response = self.generate()
                curr_iter += 1

        return response

    def call_function(self, tool_use: ToolUse) -> ToolResult:
        """Call the function indicated by the tool use"""
        tool = self._toolmap.get(tool_use.name)

        if tool is None or tool_use.is_error:
            output = f"ERROR: Failed to use tool {tool_use.id}.\nDo NOT use the same tool name and parameters again - that will lead to the same error."  # noqa: E501

            if tool_use.is_error:
                output += f"\n{tool_use.error_message}"
            elif tool is None:
                valid_tool_names = ", ".join(self._toolmap.keys())
                output += f"\nNo tool exists with the name '{tool_use.name}'. Valid tool names are: {valid_tool_names}"

            return ToolResult(tool_use_id=tool_use.id, output=output, is_error=True)

        try:
            if isinstance(tool_use.parameters, dict):
                output = json.dumps(tool.function(**tool_use.parameters))
            elif isinstance(tool_use.parameters, list):
                output = json.dumps(tool.function(*tool_use.parameters))
            else:
                raise ValueError(
                    f"The provided tool parameters, {tool_use.parameters} could not be interpreted as a mapping of arguments."  # noqa: E501
                )

            validate_tool_output(output)

            is_error = False
        except Exception as e:
            tb = traceback.format_exc()
            output = str(tb) + "\n" + str(e)
            is_error = True

        return ToolResult(tool_use_id=tool_use.id, output=output, is_error=is_error)

    def add_tool_use(self, tool_use: ToolUse) -> None:
        """Manually add a tool use and corresponding result

        This will call the implied function and add an assistant
        message requesting the ToolUse and a user message with the ToolResult
        """
        tool_result = self.call_function(tool_use)
        self.add(Message(role="assistant", content=[tool_use]))
        self.add(Message(role="user", content=[tool_result]))

    def add_checkpoints_from_usage(self, usage: Usage) -> None:
        """
        Add checkpoints to the exchange based on the token counts of the last two
        groups of messages, as well as the current token total count of the exchange
        """
        # we know we just appended one message as the response from the LLM
        # so we need to create two checkpoints as we know the token counts
        # of the last two groups of messages:
        # 1. from the last checkpoint to the most recent user message
        # 2. the most recent assistant message
        last_checkpoint_end_index = (
            self.checkpoint_data.checkpoints[-1].end_index - self.checkpoint_data.message_index_offset
            if len(self.checkpoint_data.checkpoints) > 0
            else -1
        )
        new_start_index = last_checkpoint_end_index + 1

        # here, our self.checkpoint_data.total_token_count is the previous total token count from the last time
        # that we performed a request. if we subtract this value from the input_tokens from our
        # latest response, we know how many tokens our **1** from above is.
        first_block_token_count = usage.input_tokens - self.checkpoint_data.total_token_count
        second_block_token_count = usage.output_tokens

        if len(self.messages) - new_start_index > 1:
            # this will occur most of the time, as we will have one new user message and one
            # new assistant message.

            self.checkpoint_data.checkpoints.append(
                Checkpoint(
                    start_index=new_start_index + self.checkpoint_data.message_index_offset,
                    # end index below is equivalent to the second last message. why? becuase
                    # the last message is the assistant message that we add below. we need to also
                    # track the token count of the user message sent.
                    end_index=len(self.messages) - 2 + self.checkpoint_data.message_index_offset,
                    token_count=first_block_token_count,
                )
            )
        self.checkpoint_data.checkpoints.append(
            Checkpoint(
                start_index=len(self.messages) - 1 + self.checkpoint_data.message_index_offset,
                end_index=len(self.messages) - 1 + self.checkpoint_data.message_index_offset,
                token_count=second_block_token_count,
            )
        )

        # TODO: check if the front of the checkpoints doesn't overlap with
        # the first message. if so, we are missing checkpoint data from
        # message[0] to message[checkpoint_data.checkpoints[0].start_index]
        # we can fill in this data by performing an extra request and doing some math
        self.checkpoint_data.total_token_count = usage.total_tokens

    def pop_last_message(self) -> Message:
        """Pop the last message from the exchange, handling checkpoints correctly"""
        if (
            len(self.checkpoint_data.checkpoints) > 0
            and self.checkpoint_data.last_message_index > len(self.messages) - 1
        ):
            raise ValueError("Our checkpoint data is out of sync with our message data")
        if (
            len(self.checkpoint_data.checkpoints) > 0
            and self.checkpoint_data.last_message_index == len(self.messages) - 1
        ):
            # remove the last checkpoint, because we no longer know the token count of it's contents.
            # note that this is not the same as reverting to the last checkpoint, as we want to
            # keep the messages from the last checkpoint. they will have a new checkpoint created for
            # them when we call generate() again
            self.checkpoint_data.pop()
        self.messages.pop()

    def pop_first_message(self) -> Message:
        """Pop the first message from the exchange, handling checkpoints correctly"""
        if len(self.messages) == 0:
            raise ValueError("There are no messages to pop")
        if len(self.checkpoint_data.checkpoints) == 0:
            raise ValueError("There must be at least one checkpoint to pop the first message")

        # get the start and end indexes of the first checkpoint, use these to remove message
        first_checkpoint = self.checkpoint_data.checkpoints[0]
        first_checkpoint_start_index = first_checkpoint.start_index - self.checkpoint_data.message_index_offset

        # check if the first message is part of the first checkpoint
        if first_checkpoint_start_index == 0:
            # remove this checkpoint, as it no longer has any messages
            self.checkpoint_data.pop(0)

        self.messages.pop(0)
        self.checkpoint_data.message_index_offset += 1

        if len(self.checkpoint_data.checkpoints) == 0:
            # we've removed all the checkpoints, so we need to reset the message index offset
            self.checkpoint_data.message_index_offset = 0

    def pop_last_checkpoint(self) -> Tuple[Checkpoint, List[Message]]:
        """
        Reverts the exchange back to the last checkpoint, removing associated messages
        """
        removed_checkpoint = self.checkpoint_data.checkpoints.pop()
        # pop messages until we reach the start of the next checkpoint
        messages = []
        while len(self.messages) > removed_checkpoint.start_index - self.checkpoint_data.message_index_offset:
            messages.append(self.messages.pop())
        return removed_checkpoint, messages

    def pop_first_checkpoint(self) -> Tuple[Checkpoint, List[Message]]:
        """
        Pop the first checkpoint from the exchange, removing associated messages
        """
        if len(self.checkpoint_data.checkpoints) == 0:
            raise ValueError("There are no checkpoints to pop")
        first_checkpoint = self.checkpoint_data.pop(0)

        # remove messages until we reach the start of the next checkpoint
        messages = []
        stop_at_index = first_checkpoint.end_index - self.checkpoint_data.message_index_offset
        for _ in range(stop_at_index + 1):  # +1 because it's inclusive
            messages.append(self.messages.pop(0))
            self.checkpoint_data.message_index_offset += 1

        if len(self.checkpoint_data.checkpoints) == 0:
            # we've removed all the checkpoints, so we need to reset the message index offset
            self.checkpoint_data.message_index_offset = 0
        return first_checkpoint, messages

    def prepend_checkpointed_message(self, message: Message, token_count: int) -> None:
        """Prepend a message to the exchange, updating the checkpoint data"""
        self.messages.insert(0, message)
        new_index = max(0, self.checkpoint_data.message_index_offset - 1)
        self.checkpoint_data.checkpoints.insert(
            0,
            Checkpoint(
                start_index=new_index,
                end_index=new_index,
                token_count=token_count,
            ),
        )
        self.checkpoint_data.message_index_offset = new_index

    def rewind(self) -> None:
        if not self.messages:
            return

        # we remove messages until we find the last user text message
        while not (self.messages[-1].role == "user" and type(self.messages[-1].content[-1]) is Text):
            self.pop_last_message()

        # now we remove that last user text message, putting us at a good point
        # to ask the user for their input again
        if self.messages:
            self.pop_last_message()

    @property
    def is_allowed_to_call_llm(self) -> bool:
        """
        Returns True if the exchange is allowed to call the LLM, False otherwise
        """
        # TODO: reconsider whether this function belongs here and whether it is necessary
        # Some models will have different requirements than others, so it may be better for
        # this to be a required method of the provider instead.
        return len(self.messages) > 0 and self.messages[-1].role == "user"

    def get_token_usage(self) -> Dict[str, Usage]:
        return _token_usage_collector.get_token_usage_group_by_model()



import os
from typing import Type

import httpx

from exchange.providers.openai import OpenAiProvider

OLLAMA_HOST = "http://localhost:11434/"
OLLAMA_MODEL = "mistral-nemo"


class OllamaProvider(OpenAiProvider):
    """Provides chat completions for models hosted by Ollama"""

    __doc__ += f"""

Here's an example profile configuration to try:

    ollama:
      provider: ollama
      processor: {OLLAMA_MODEL}
      accelerator: {OLLAMA_MODEL}
      moderator: passive
      toolkits:
      - name: developer
        requires: {{}}
"""

    def __init__(self, client: httpx.Client) -> None:
        print("PLEASE NOTE: the ollama provider is experimental, use with care")
        super().__init__(client)

    @classmethod
    def from_env(cls: Type["OllamaProvider"]) -> "OllamaProvider":
        url = os.environ.get("OLLAMA_HOST", OLLAMA_HOST)
        client = httpx.Client(
            base_url=url,
            timeout=httpx.Timeout(60 * 10),
        )
        # from_env is expected to fail if required ENV variables are not
        # available. Since this provider can run with defaults, we substitute
        # a health check to verify the endpoint is running.
        client.get("")
        # The OpenAI API is defined after "v1/", so we need to join it here.
        client.base_url = client.base_url.join("v1/")
        return cls(client)



from unittest.mock import MagicMock
from exchange.exchange import Exchange
from exchange.message import Message
from exchange.moderators.passive import PassiveModerator
from exchange.providers.base import Provider
from exchange.tool import Tool
from exchange.token_usage_collector import _TokenUsageCollector

MODEL_NAME = "test-model"


def create_exchange(mock_provider, dummy_tool):
    return Exchange(
        provider=mock_provider,
        model=MODEL_NAME,
        system="test-system",
        tools=(Tool.from_function(dummy_tool),),
        messages=[],
        moderator=PassiveModerator(),
    )


def test_exchange_generate_collect_usage(usage_factory, dummy_tool, monkeypatch):
    mock_provider = MagicMock(spec=Provider)
    mock_usage_collector = MagicMock(spec=_TokenUsageCollector)
    usage = usage_factory()
    mock_provider.complete.return_value = (Message.assistant("msg"), usage)
    exchange = create_exchange(mock_provider, dummy_tool)

    monkeypatch.setattr("exchange.exchange._token_usage_collector", mock_usage_collector)
    exchange.generate()

    mock_usage_collector.collect.assert_called_once_with(MODEL_NAME, usage)



import pytest

from exchange.providers.base import Usage


@pytest.fixture
def dummy_tool():
    def _dummy_tool() -> str:
        """An example tool"""
        return "dummy response"

    return _dummy_tool


@pytest.fixture
def usage_factory():
    def _create_usage(input_tokens=100, output_tokens=200, total_tokens=300):
        return Usage(input_tokens=input_tokens, output_tokens=output_tokens, total_tokens=total_tokens)

    return _create_usage


def read_file(filename: str) -> str:
    """
    Read the contents of the file.

    Args:
        filename (str): The path to the file, which can be relative or
            absolute. If it is a plain filename, it is assumed to be in the
            current working directory.

    Returns:
        str: The contents of the file.
    """
    assert filename == "test.txt"
    return "hello exchange"



import os
from typing import Type, Tuple

import pytest

from exchange import Message, ToolUse, ToolResult, Tool
from exchange.providers import Usage, Provider
from tests.conftest import read_file

OPENAI_API_KEY = "test_openai_api_key"
OPENAI_ORG_ID = "test_openai_org_key"
OPENAI_PROJECT_ID = "test_openai_project_id"


@pytest.fixture
def default_openai_env(monkeypatch):
    """
    This fixture prevents OpenAIProvider.from_env() from erring on missing
    environment variables.

    When running VCR tests for the first time or after deleting a cassette
    recording, set required environment variables, so that real requests don't
    fail. Subsequent runs use the recorded data, so don't need them.
    """
    if "OPENAI_API_KEY" not in os.environ:
        monkeypatch.setenv("OPENAI_API_KEY", OPENAI_API_KEY)


@pytest.fixture(scope="module")
def vcr_config():
    """
    This scrubs sensitive data and gunzips bodies when in recording mode.

    Without this, you would leak cookies and auth tokens in the cassettes.
    Also, depending on the request, some responses would be binary encoded
    while others plain json. This ensures all bodies are human-readable.
    """
    return {
        "decode_compressed_response": True,
        "filter_headers": [
            ("authorization", "Bearer " + OPENAI_API_KEY),
            ("openai-organization", OPENAI_ORG_ID),
            ("openai-project", OPENAI_PROJECT_ID),
            ("cookie", None),
        ],
        "before_record_response": scrub_response_headers,
    }


def scrub_response_headers(response):
    """
    This scrubs sensitive response headers. Note they are case-sensitive!
    """
    response["headers"]["openai-organization"] = OPENAI_ORG_ID
    response["headers"]["Set-Cookie"] = "test_set_cookie"
    return response


def complete(provider_cls: Type[Provider], model: str) -> Tuple[Message, Usage]:
    provider = provider_cls.from_env()
    system = "You are a helpful assistant."
    messages = [Message.user("Hello")]
    return provider.complete(model=model, system=system, messages=messages, tools=None)


def tools(provider_cls: Type[Provider], model: str) -> Tuple[Message, Usage]:
    provider = provider_cls.from_env()
    system = "You are a helpful assistant. Expect to need to read a file using read_file."
    messages = [Message.user("What are the contents of this file? test.txt")]
    return provider.complete(model=model, system=system, messages=messages, tools=(Tool.from_function(read_file),))


def vision(provider_cls: Type[Provider], model: str) -> Tuple[Message, Usage]:
    provider = provider_cls.from_env()
    system = "You are a helpful assistant."
    messages = [
        Message.user("What does the first entry in the menu say?"),
        Message(
            role="assistant",
            content=[ToolUse(id="xyz", name="screenshot", parameters={})],
        ),
        Message(
            role="user",
            content=[ToolResult(tool_use_id="xyz", output='"image:tests/test_image.png"')],
        ),
    ]
    return provider.complete(model=model, system=system, messages=messages, tools=None)



import pytest
from exchange import Exchange
from exchange.content import ToolResult, ToolUse
from exchange.message import Message
from exchange.moderators.truncate import ContextTruncate
from exchange.providers import Provider, Usage

MAX_TOKENS = 300
SYSTEM_PROMPT_TOKENS = 100

MESSAGE_SEQUENCE = [
    Message.user("Hi, can you help me with my homework?"),
    Message.assistant("Of course! What do you need help with?"),
    Message.user("I need help with math problems."),
    Message.assistant("Sure, I can help with that. Let's get started."),
    Message.user("What is 2 + 2, 3*3, 9/5, 2*20, 14/2?"),
    Message(
        role="assistant",
        content=[ToolUse(id="1", name="add", parameters={"a": 2, "b": 2})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="1", output="4")]),
    Message(
        role="assistant",
        content=[ToolUse(id="2", name="multiply", parameters={"a": 3, "b": 3})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="2", output="9")]),
    Message(
        role="assistant",
        content=[ToolUse(id="3", name="divide", parameters={"a": 9, "b": 5})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="3", output="1.8")]),
    Message(
        role="assistant",
        content=[ToolUse(id="4", name="multiply", parameters={"a": 2, "b": 20})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="4", output="40")]),
    Message(
        role="assistant",
        content=[ToolUse(id="5", name="divide", parameters={"a": 14, "b": 2})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="5", output="7")]),
    Message.assistant("I'm done calculating the answers to your math questions."),
    Message.user("Can you also help with my science homework?"),
    Message.assistant("Yes, I can help with science too."),
    Message.user("What is the speed of light? The frequency of a photon? The mass of an electron?"),
    Message(
        role="assistant",
        content=[ToolUse(id="6", name="speed_of_light", parameters={})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="6", output="299,792,458 m/s")]),
    Message(
        role="assistant",
        content=[ToolUse(id="7", name="photon_frequency", parameters={})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="7", output="2.418 x 10^14 Hz")]),
    Message(role="assistant", content=[ToolUse(id="8", name="electron_mass", parameters={})]),
    Message(
        role="user",
        content=[ToolResult(tool_use_id="8", output="9.10938356 x 10^-31 kg")],
    ),
    Message.assistant("I'm done calculating the answers to your science questions."),
    Message.user("That's great! How about history?"),
    Message.assistant("Of course, I can help with history as well."),
    Message.user("Thanks! You're very helpful."),
    Message.assistant("You're welcome! I'm here to help."),
]


class TruncateLinearProvider(Provider):
    def __init__(self):
        self.sequence = MESSAGE_SEQUENCE
        self.current_index = 1
        self.summarize_next = False
        self.summarized_count = 0

    def complete(self, model, system, messages, tools):
        input_token_count = SYSTEM_PROMPT_TOKENS

        message = self.sequence[self.current_index]

        if len(messages) > 0 and type(messages[0].content[0]) is ToolResult:
            raise ValueError("ToolResult should not be the first message")

        if len(messages) == 1 and messages[0].text == "a":
            # adding a +1 for the "a"
            return Message.assistant("Getting system prompt size"), Usage(
                input_tokens=80 + 1, output_tokens=20, total_tokens=SYSTEM_PROMPT_TOKENS + 1
            )

        for i in range(len(messages)):
            if type(messages[i].content[0]) in (ToolResult, ToolUse):
                input_token_count += 10
            else:
                input_token_count += len(messages[i].text) * 2

        if type(message.content[0]) in (ToolResult, ToolUse):
            output_tokens = 10
        else:
            output_tokens = len(message.text) * 2

        total_tokens = input_token_count + output_tokens
        usage = Usage(
            input_tokens=input_token_count,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )
        self.current_index += 2
        return message, usage


@pytest.fixture
def conversation_exchange_instance():
    ex = Exchange(
        provider=TruncateLinearProvider(),
        model="test-model",
        system="test-system",
        moderator=ContextTruncate(max_tokens=500),
    )
    return ex


def test_truncate_on_generic_conversation(conversation_exchange_instance: Exchange):
    i = 0
    while i < len(MESSAGE_SEQUENCE):
        next_message = MESSAGE_SEQUENCE[i]
        conversation_exchange_instance.add(next_message)
        message = conversation_exchange_instance.generate()
        if message.text != "Summary message here":
            i += 2
        # ensure the total token count is not anything exhorbitant
        assert conversation_exchange_instance.checkpoint_data.total_token_count < 700
        assert conversation_exchange_instance.moderator.system_prompt_token_count == 100



from typing import Type

from exchange import Message
from exchange.checkpoint import CheckpointData
from exchange.moderators import ContextTruncate, PassiveModerator


class ContextSummarizer(ContextTruncate):
    def rewrite(self, exchange: Type["exchange.exchange.Exchange"]) -> None:  # noqa: F821
        """Summarize the context history up to the last few messages in the exchange"""

        self._update_system_prompt_token_count(exchange)

        # TODO: use an offset for summarization
        if exchange.checkpoint_data.total_token_count < self.max_tokens:
            return

        messages_to_summarize = self._get_messages_to_remove(exchange)
        num_messages_to_remove = len(messages_to_summarize)

        # the llm will throw an error if the last message isn't a user message
        if messages_to_summarize[-1].role == "assistant" and (not messages_to_summarize[-1].tool_use):
            messages_to_summarize.append(Message.user("Summarize our the above conversation"))

        summarizer_exchange = exchange.replace(
            system=Message.load("summarizer.jinja").text,
            moderator=PassiveModerator(),
            model=self.model,
            messages=messages_to_summarize,
            checkpoint_data=CheckpointData(),
        )

        # get the summarized content and the tokens associated with this content
        summary = summarizer_exchange.reply()
        summary_checkpoint = summarizer_exchange.checkpoint_data.checkpoints[-1]

        # remove the checkpoints that were summarized from the original exchange
        for _ in range(num_messages_to_remove):
            exchange.pop_first_message()

        # insert summary as first message/checkpoint
        if len(exchange.messages) == 0 or exchange.messages[0].role == "assistant":
            summary_message = Message.user(summary.text)
        else:
            summary_message = Message.assistant(summary.text)
        exchange.prepend_checkpointed_message(summary_message, summary_checkpoint.token_count)



from exchange.token_usage_collector import _TokenUsageCollector


def test_collect(usage_factory):
    usage_collector = _TokenUsageCollector()
    usage_collector.collect("model1", usage_factory(100, 1000, 1100))
    usage_collector.collect("model1", usage_factory(200, 2000, 2200))
    usage_collector.collect("model2", usage_factory(400, 4000, 4400))
    usage_collector.collect("model3", usage_factory(500, 5000, 5500))
    usage_collector.collect("model3", usage_factory(600, 6000, 6600))
    assert usage_collector.get_token_usage_group_by_model() == {
        "model1": usage_factory(300, 3000, 3300),
        "model2": usage_factory(400, 4000, 4400),
        "model3": usage_factory(1100, 11000, 12100),
    }


def test_collect_with_non_input_or_output_token(usage_factory):
    usage_collector = _TokenUsageCollector()
    usage_collector.collect("model1", usage_factory(100, None, None))
    usage_collector.collect("model1", usage_factory(None, 2000, None))
    assert usage_collector.get_token_usage_group_by_model() == {
        "model1": usage_factory(100, 2000, 0),
    }



import os
import pytest
from exchange.exchange import Exchange
from exchange.message import Message
from exchange.moderators import ContextTruncate
from exchange.providers import get_provider
from exchange.providers.ollama import OLLAMA_MODEL
from exchange.tool import Tool
from tests.conftest import read_file

too_long_chars = "x" * (2**20 + 1)

cases = [
    # Set seed and temperature for more determinism, to avoid flakes
    (get_provider("ollama"), os.getenv("OLLAMA_MODEL", OLLAMA_MODEL), dict(seed=3, temperature=0.1)),
    (get_provider("openai"), os.getenv("OPENAI_MODEL", "gpt-4o-mini"), dict()),
    (get_provider("databricks"), "databricks-meta-llama-3-70b-instruct", dict()),
    (get_provider("bedrock"), "anthropic.claude-3-5-sonnet-20240620-v1:0", dict()),
]


@pytest.mark.integration
@pytest.mark.parametrize("provider,model,kwargs", cases)
def test_simple(provider, model, kwargs):
    provider = provider.from_env()

    ex = Exchange(
        provider=provider,
        model=model,
        moderator=ContextTruncate(model),
        system="You are a helpful assistant.",
        generation_args=kwargs,
    )

    ex.add(Message.user("Who is the most famous wizard from the lord of the rings"))

    response = ex.reply()

    # It's possible this can be flakey, but in experience so far haven't seen it
    assert "gandalf" in response.text.lower()


@pytest.mark.integration
@pytest.mark.parametrize("provider,model,kwargs", cases)
def test_tools(provider, model, kwargs, tmp_path):
    provider = provider.from_env()

    ex = Exchange(
        provider=provider,
        model=model,
        moderator=ContextTruncate(model),
        system="You are a helpful assistant. Expect to need to read a file using read_file.",
        tools=(Tool.from_function(read_file),),
        generation_args=kwargs,
    )

    ex.add(Message.user("What are the contents of this file? test.txt"))

    response = ex.reply()

    assert "hello exchange" in response.text.lower()


@pytest.mark.integration
@pytest.mark.parametrize("provider,model,kwargs", cases)
def test_tool_use_output_chars(provider, model, kwargs):
    provider = provider.from_env()

    def get_password() -> str:
        """Return the password for authentication"""
        return too_long_chars

    ex = Exchange(
        provider=provider,
        model=model,
        moderator=ContextTruncate(model),
        system="You are a helpful assistant. Expect to need to authenticate using get_password.",
        tools=(Tool.from_function(get_password),),
        generation_args=kwargs,
    )

    ex.add(Message.user("Can you authenticate this session by responding with the password"))

    ex.reply()

    # Without our error handling, this would raise
    # string too long. Expected a string with maximum length 1048576, but got a string with length ...



import os
from typing import Any, Dict, List, Tuple, Type

import httpx

from exchange.message import Message
from exchange.providers.base import Provider, Usage
from tenacity import retry, wait_fixed, stop_after_attempt
from exchange.providers.utils import retry_if_status
from exchange.providers.utils import (
    messages_to_openai_spec,
    openai_response_to_message,
    openai_single_message_context_length_exceeded,
    raise_for_status,
    tools_to_openai_spec,
)
from exchange.tool import Tool

retry_procedure = retry(
    wait=wait_fixed(2),
    stop=stop_after_attempt(2),
    retry=retry_if_status(codes=[429], above=500),
    reraise=True,
)


class AzureProvider(Provider):
    """Provides chat completions for models hosted directly by OpenAI"""

    def __init__(self, client: httpx.Client, deployment_name: str, api_version: str) -> None:
        super().__init__()
        self.client = client
        self.deployment_name = deployment_name
        self.api_version = api_version

    @classmethod
    def from_env(cls: Type["AzureProvider"]) -> "AzureProvider":
        try:
            url = os.environ["AZURE_CHAT_COMPLETIONS_HOST_NAME"]
        except KeyError:
            raise RuntimeError("Failed to get AZURE_CHAT_COMPLETIONS_HOST_NAME from the environment.")

        try:
            deployment_name = os.environ["AZURE_CHAT_COMPLETIONS_DEPLOYMENT_NAME"]
        except KeyError:
            raise RuntimeError("Failed to get AZURE_CHAT_COMPLETIONS_DEPLOYMENT_NAME from the environment.")

        try:
            api_version = os.environ["AZURE_CHAT_COMPLETIONS_DEPLOYMENT_API_VERSION"]
        except KeyError:
            raise RuntimeError("Failed to get AZURE_CHAT_COMPLETIONS_DEPLOYMENT_API_VERSION from the environment.")

        try:
            key = os.environ["AZURE_CHAT_COMPLETIONS_KEY"]
        except KeyError:
            raise RuntimeError("Failed to get AZURE_CHAT_COMPLETIONS_KEY from the environment.")

        # format the url host/"openai/deployments/" + deployment_name + "/chat/completions?api-version=" + api_version
        url = f"{url}/openai/deployments/{deployment_name}"
        client = httpx.Client(
            base_url=url,
            headers={"api-key": key, "Content-Type": "application/json"},
            timeout=httpx.Timeout(60 * 10),
        )
        return cls(client, deployment_name, api_version)

    @staticmethod
    def get_usage(data: dict) -> Usage:
        usage = data.pop("usage")
        input_tokens = usage.get("prompt_tokens")
        output_tokens = usage.get("completion_tokens")
        total_tokens = usage.get("total_tokens")

        if total_tokens is None and input_tokens is not None and output_tokens is not None:
            total_tokens = input_tokens + output_tokens

        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )

    def complete(
        self,
        model: str,
        system: str,
        messages: List[Message],
        tools: Tuple[Tool],
        **kwargs: Dict[str, Any],
    ) -> Tuple[Message, Usage]:
        payload = dict(
            messages=[
                {"role": "system", "content": system},
                *messages_to_openai_spec(messages),
            ],
            tools=tools_to_openai_spec(tools) if tools else [],
            **kwargs,
        )

        payload = {k: v for k, v in payload.items() if v}
        request_url = f"{self.client.base_url}/chat/completions?api-version={self.api_version}"
        response = self._post(payload, request_url)

        # Check for context_length_exceeded error for single, long input message
        if "error" in response and len(messages) == 1:
            openai_single_message_context_length_exceeded(response["error"])

        message = openai_response_to_message(response)
        usage = self.get_usage(response)
        return message, usage

    @retry_procedure
    def _post(self, payload: dict, request_url: str) -> dict:
        response = self.client.post(request_url, json=payload)
        return raise_for_status(response).json()



from copy import deepcopy
from typing import List
from attrs import define, field


@define
class Checkpoint:
    """Checkpoint that counts the tokens in messages between the start and end index"""

    start_index: int = field(default=0)  # inclusive
    end_index: int = field(default=0)  # inclusive
    token_count: int = field(default=0)

    def __deepcopy__(self, _) -> "Checkpoint":  # noqa: ANN001
        """
        Returns a deep copy of the Checkpoint object.
        """
        return Checkpoint(
            start_index=self.start_index,
            end_index=self.end_index,
            token_count=self.token_count,
        )


@define
class CheckpointData:
    """Aggregates all information about checkpoints"""

    # the total number of tokens in the exchange. this is updated every time a checkpoint is
    # added or removed
    total_token_count: int = field(default=0)

    # in order list of individual checkpoints in the exchange
    checkpoints: List[Checkpoint] = field(factory=list)

    # the offset to apply to the message index when calculating the last message index
    # this is useful because messages on the exchange behave like a queue, where you can only
    # pop from the left or right sides. This offset allows us to map the checkpoint indices
    # to the correct message index, even if we have popped messages from the left side of
    # the exchange in the past. we reset this offset to 0 when we empty the checkpoint data.
    message_index_offset: int = field(default=0)

    def __deepcopy__(self, memo: dict) -> "CheckpointData":
        """Returns a deep copy of the CheckpointData object."""
        return CheckpointData(
            total_token_count=self.total_token_count,
            checkpoints=deepcopy(self.checkpoints, memo),
            message_index_offset=self.message_index_offset,
        )

    @property
    def last_message_index(self) -> int:
        if not self.checkpoints:
            return -1  # we don't have enough information to know
        return self.checkpoints[-1].end_index - self.message_index_offset

    def reset(self) -> None:
        """Resets the checkpoint data to its initial state."""
        self.checkpoints = []
        self.message_index_offset = 0
        self.total_token_count = 0

    def pop(self, index: int = -1) -> Checkpoint:
        """Removes and returns the checkpoint at the given index."""
        popped_checkpoint = self.checkpoints.pop(index)
        self.total_token_count = self.total_token_count - popped_checkpoint.token_count
        return popped_checkpoint



import os
from unittest.mock import patch

import pytest
from exchange import Message, Text
from exchange.providers.azure import AzureProvider


@pytest.fixture
@patch.dict(
    os.environ,
    {
        "AZURE_CHAT_COMPLETIONS_HOST_NAME": "https://test.openai.azure.com/",
        "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_NAME": "test-deployment",
        "AZURE_CHAT_COMPLETIONS_DEPLOYMENT_API_VERSION": "2024-02-15-preview",
        "AZURE_CHAT_COMPLETIONS_KEY": "test_api_key",
    },
)
def azure_provider():
    return AzureProvider.from_env()


@patch("httpx.Client.post")
@patch("time.sleep", return_value=None)
@patch("logging.warning")
@patch("logging.error")
def test_azure_completion(mock_error, mock_warning, mock_sleep, mock_post, azure_provider):
    mock_response = {
        "choices": [{"message": {"role": "assistant", "content": "Hello!"}}],
        "usage": {"prompt_tokens": 10, "completion_tokens": 25, "total_tokens": 35},
    }

    mock_post.return_value.json.return_value = mock_response

    model = "gpt-4"
    system = "You are a helpful assistant."
    messages = [Message.user("Hello")]
    tools = ()

    reply_message, reply_usage = azure_provider.complete(model=model, system=system, messages=messages, tools=tools)

    assert reply_message.content == [Text(text="Hello!")]
    assert reply_usage.total_tokens == 35
    mock_post.assert_called_once_with(
        f"{azure_provider.client.base_url}/chat/completions?api-version={azure_provider.api_version}",
        json={
            "messages": [
                {"role": "system", "content": system},
                {"role": "user", "content": "Hello"},
            ],
        },
    )


@pytest.mark.integration
def test_azure_integration():
    provider = AzureProvider.from_env()
    system = "You are a helpful assistant."
    messages = [Message.user("Hello")]

    reply = provider.complete(system=system, messages=messages, tools=None)

    assert reply[0].content is not None
    print("Completion content from Azure:", reply[0].content)



import os

import pytest

from exchange import Text, ToolUse
from exchange.providers.ollama import OllamaProvider, OLLAMA_MODEL
from .conftest import complete, tools

OLLAMA_MODEL = os.getenv("OLLAMA_MODEL", OLLAMA_MODEL)


@pytest.mark.vcr()
def test_ollama_complete():
    reply_message, reply_usage = complete(OllamaProvider, OLLAMA_MODEL)

    assert reply_message.content == [Text(text="Hello! I'm here to help. How can I assist you today? Let's chat. 😊")]
    assert reply_usage.total_tokens == 33


@pytest.mark.integration
def test_ollama_complete_integration():
    reply = complete(OllamaProvider, OLLAMA_MODEL)

    assert reply[0].content is not None
    print("Completion content from OpenAI:", reply[0].content)


@pytest.mark.vcr()
def test_ollama_tools():
    reply_message, reply_usage = tools(OllamaProvider, OLLAMA_MODEL)

    tool_use = reply_message.content[0]
    assert isinstance(tool_use, ToolUse)
    assert tool_use.id == "call_z6fgu3z3"
    assert tool_use.name == "read_file"
    assert tool_use.parameters == {"filename": "test.txt"}
    assert reply_usage.total_tokens == 133


@pytest.mark.integration
def test_ollama_tools_integration():
    reply = tools(OllamaProvider, OLLAMA_MODEL)

    tool_use = reply[0].content[0]
    assert isinstance(tool_use, ToolUse)
    assert tool_use.id is not None
    assert tool_use.name == "read_file"
    assert tool_use.parameters == {"filename": "test.txt"}



from typing import List, Tuple

import pytest

from exchange.checkpoint import Checkpoint, CheckpointData
from exchange.content import Text, ToolResult, ToolUse
from exchange.exchange import Exchange
from exchange.message import Message
from exchange.moderators import PassiveModerator
from exchange.providers import Provider, Usage
from exchange.tool import Tool


def dummy_tool() -> str:
    """An example tool"""
    return "dummy response"


too_long_output = "x" * (2**20 + 1)
too_long_token_output = "word " * 128000


def no_overlapping_checkpoints(exchange: Exchange) -> bool:
    """Assert that there are no overlapping checkpoints in the exchange."""
    for i, checkpoint in enumerate(exchange.checkpoint_data.checkpoints):
        for other_checkpoint in exchange.checkpoint_data.checkpoints[i + 1 :]:
            if not checkpoint.end_index < other_checkpoint.start_index:
                return False
    return True


def checkpoint_to_index_pairs(checkpoints: List[Checkpoint]) -> List[Tuple[int, int]]:
    return [(checkpoint.start_index, checkpoint.end_index) for checkpoint in checkpoints]


class MockProvider(Provider):
    def __init__(self, sequence: List[Message], usage_dicts: List[dict]):
        # We'll use init to provide a preplanned reply sequence
        self.sequence = sequence
        self.call_count = 0
        self.usage_dicts = usage_dicts

    @staticmethod
    def get_usage(data: dict) -> Usage:
        usage = data.pop("usage")
        input_tokens = usage.get("input_tokens")
        output_tokens = usage.get("output_tokens")
        total_tokens = usage.get("total_tokens")

        if total_tokens is None and input_tokens is not None and output_tokens is not None:
            total_tokens = input_tokens + output_tokens

        return Usage(
            input_tokens=input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )

    def complete(self, model: str, system: str, messages: List[Message], tools: List[Tool]) -> Message:
        output = self.sequence[self.call_count]
        usage = self.get_usage(self.usage_dicts[self.call_count])
        self.call_count += 1
        return (output, usage)


def test_reply_with_unsupported_tool():
    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="unsupported_tool", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[Text(text="Here is the completion after tool call")],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=(Tool.from_function(dummy_tool),),
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test")]))

    ex.reply()

    content = ex.messages[-2].content[0]
    assert isinstance(content, ToolResult) and content.is_error and "no tool exists" in content.output.lower()


def test_invalid_tool_parameters():
    """Test handling of invalid tool parameters response"""
    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="dummy_tool", parameters="invalid json")],
                ),
                Message(
                    role="assistant",
                    content=[Text(text="Here is the completion after tool call")],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(dummy_tool)],
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test invalid parameters")]))

    ex.reply()

    content = ex.messages[-2].content[0]
    assert isinstance(content, ToolResult) and content.is_error and "invalid json" in content.output.lower()


def test_max_tool_use_when_limit_reached():
    """Test the max_tool_use parameter in the reply method."""
    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="dummy_tool", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[ToolUse(id="2", name="dummy_tool", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[ToolUse(id="3", name="dummy_tool", parameters={})],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(dummy_tool)],
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test max tool use")]))

    response = ex.reply(max_tool_use=3)

    assert ex.provider.call_count == 3
    assert "reached the limit" in response.content[0].text.lower()

    assert isinstance(ex.messages[-2].content[0], ToolResult) and ex.messages[-2].content[0].tool_use_id == "3"

    assert ex.messages[-1].role == "assistant"


def test_tool_output_too_long_character_error():
    """Test tool handling when output exceeds character limit."""

    def long_output_tool_char() -> str:
        return too_long_output

    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="long_output_tool_char", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[Text(text="Here is the completion after tool call")],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(long_output_tool_char)],
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test long output char")]))

    ex.reply()

    content = ex.messages[-2].content[0]
    assert (
        isinstance(content, ToolResult)
        and content.is_error
        and "output that was too long to handle" in content.output.lower()
    )


def test_tool_output_too_long_token_error():
    """Test tool handling when output exceeds token limit."""

    def long_output_tool_token() -> str:
        return too_long_token_output

    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="long_output_tool_token", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[Text(text="Here is the completion after tool call")],
                ),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(long_output_tool_token)],
        moderator=PassiveModerator(),
    )

    ex.add(Message(role="user", content=[Text(text="test long output token")]))

    ex.reply()

    content = ex.messages[-2].content[0]
    assert (
        isinstance(content, ToolResult)
        and content.is_error
        and "output that was too long to handle" in content.output.lower()
    )


@pytest.fixture(scope="function")
def normal_exchange() -> Exchange:
    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message(role="assistant", content=[Text(text="Message 1")]),
                Message(role="assistant", content=[Text(text="Message 2")]),
                Message(role="assistant", content=[Text(text="Message 3")]),
                Message(role="assistant", content=[Text(text="Message 4")]),
                Message(role="assistant", content=[Text(text="Message 5")]),
            ],
            usage_dicts=[
                {"usage": {"total_tokens": 10, "input_tokens": 5, "output_tokens": 5}},
                {"usage": {"total_tokens": 28, "input_tokens": 10, "output_tokens": 18}},
                {"usage": {"total_tokens": 33, "input_tokens": 28, "output_tokens": 5}},
                {"usage": {"total_tokens": 40, "input_tokens": 32, "output_tokens": 8}},
                {"usage": {"total_tokens": 50, "input_tokens": 40, "output_tokens": 10}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=(Tool.from_function(dummy_tool),),
        moderator=PassiveModerator(),
        checkpoint_data=CheckpointData(),
    )
    return ex


@pytest.fixture(scope="function")
def resumed_exchange() -> Exchange:
    messages = [
        Message(role="user", content=[Text(text="User message 1")]),
        Message(role="assistant", content=[Text(text="Assistant Message 1")]),
        Message(role="user", content=[Text(text="User message 2")]),
        Message(role="assistant", content=[Text(text="Assistant Message 2")]),
        Message(role="user", content=[Text(text="User message 3")]),
        Message(role="assistant", content=[Text(text="Assistant Message 3")]),
    ]
    provider = MockProvider(
        sequence=[
            Message(role="assistant", content=[Text(text="Assistant Message 4")]),
        ],
        usage_dicts=[
            {"usage": {"total_tokens": 40, "input_tokens": 32, "output_tokens": 8}},
        ],
    )
    ex = Exchange(
        provider=provider,
        messages=messages,
        tools=[],
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        checkpoint_data=CheckpointData(),
        moderator=PassiveModerator(),
    )
    return ex


def test_checkpoints_on_exchange(normal_exchange):
    """Test checkpoints on an exchange."""
    ex = normal_exchange
    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()
    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()
    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()

    # Check if checkpoints are created correctly
    checkpoints = ex.checkpoint_data.checkpoints
    assert len(checkpoints) == 6
    for i in range(len(ex.messages)):
        # asserting that each message has a corresponding checkpoint
        assert checkpoints[i].start_index == i
        assert checkpoints[i].end_index == i

    # Check if the messages are ordered correctly
    assert [msg.content[0].text for msg in ex.messages] == [
        "User message",
        "Message 1",
        "User message",
        "Message 2",
        "User message",
        "Message 3",
    ]
    assert no_overlapping_checkpoints(ex)


def test_checkpoints_on_resumed_exchange(resumed_exchange) -> None:
    ex = resumed_exchange
    ex.pop_last_message()
    ex.reply()

    checkpoints = ex.checkpoint_data.checkpoints
    assert len(checkpoints) == 2
    assert len(ex.messages) == 6
    assert checkpoints[0].token_count == 32
    assert checkpoints[0].start_index == 0
    assert checkpoints[0].end_index == 4
    assert checkpoints[1].token_count == 8
    assert checkpoints[1].start_index == 5
    assert checkpoints[1].end_index == 5
    assert no_overlapping_checkpoints(ex)


def test_pop_last_checkpoint_on_resumed_exchange(resumed_exchange) -> None:
    ex = resumed_exchange
    ex.add(Message(role="user", content=[Text(text="Assistant Message 4")]))
    ex.reply()
    ex.pop_last_checkpoint()

    assert len(ex.messages) == 7
    assert len(ex.checkpoint_data.checkpoints) == 1

    ex.pop_last_checkpoint()
    assert len(ex.messages) == 0
    assert len(ex.checkpoint_data.checkpoints) == 0
    assert no_overlapping_checkpoints(ex)


def test_pop_last_checkpoint_on_normal_exchange(normal_exchange) -> None:
    ex = normal_exchange
    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()
    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()
    ex.pop_last_checkpoint()
    ex.pop_last_checkpoint()

    assert len(ex.messages) == 2
    assert len(ex.checkpoint_data.checkpoints) == 2
    assert no_overlapping_checkpoints(ex)
    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.pop_last_checkpoint()
    assert len(ex.messages) == 1
    assert len(ex.checkpoint_data.checkpoints) == 1
    ex.reply()
    assert len(ex.messages) == 2
    assert len(ex.checkpoint_data.checkpoints) == 2
    assert no_overlapping_checkpoints(ex)


def test_pop_first_message_no_messages():
    ex = Exchange(
        provider=MockProvider(sequence=[], usage_dicts=[]),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(dummy_tool)],
        moderator=PassiveModerator(),
    )

    with pytest.raises(ValueError) as e:
        ex.pop_first_message()
    assert str(e.value) == "There are no messages to pop"


def test_pop_first_message_checkpoint_with_many_messages(resumed_exchange):
    ex = resumed_exchange
    ex.pop_last_message()
    ex.reply()

    assert len(ex.messages) == 6
    assert len(ex.checkpoint_data.checkpoints) == 2
    assert ex.checkpoint_data.checkpoints[0].start_index == 0
    assert ex.checkpoint_data.checkpoints[0].end_index == 4
    assert ex.checkpoint_data.checkpoints[1].start_index == 5
    assert ex.checkpoint_data.checkpoints[1].end_index == 5
    assert ex.checkpoint_data.message_index_offset == 0
    assert no_overlapping_checkpoints(ex)

    ex.pop_first_message()

    assert len(ex.messages) == 5
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert ex.checkpoint_data.checkpoints[0].start_index == 5
    assert ex.checkpoint_data.checkpoints[0].end_index == 5
    assert ex.checkpoint_data.message_index_offset == 1
    assert no_overlapping_checkpoints(ex)

    ex.pop_first_message()

    assert len(ex.messages) == 4
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert ex.checkpoint_data.checkpoints[0].start_index == 5
    assert ex.checkpoint_data.checkpoints[0].end_index == 5
    assert ex.checkpoint_data.message_index_offset == 2
    assert no_overlapping_checkpoints(ex)

    ex.pop_first_message()

    assert len(ex.messages) == 3
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert ex.checkpoint_data.checkpoints[0].start_index == 5
    assert ex.checkpoint_data.checkpoints[0].end_index == 5
    assert ex.checkpoint_data.message_index_offset == 3
    assert no_overlapping_checkpoints(ex)

    ex.pop_first_message()

    assert len(ex.messages) == 2
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert ex.checkpoint_data.checkpoints[0].start_index == 5
    assert ex.checkpoint_data.checkpoints[0].end_index == 5
    assert ex.checkpoint_data.message_index_offset == 4
    assert no_overlapping_checkpoints(ex)

    ex.pop_first_message()

    assert len(ex.messages) == 1
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert ex.checkpoint_data.checkpoints[0].start_index == 5
    assert ex.checkpoint_data.checkpoints[0].end_index == 5
    assert ex.checkpoint_data.message_index_offset == 5
    assert no_overlapping_checkpoints(ex)

    ex.pop_first_message()

    assert len(ex.messages) == 0
    assert len(ex.checkpoint_data.checkpoints) == 0
    assert ex.checkpoint_data.message_index_offset == 0
    assert no_overlapping_checkpoints(ex)

    with pytest.raises(ValueError) as e:
        ex.pop_first_message()

    assert str(e.value) == "There are no messages to pop"


def test_varied_message_manipulation(normal_exchange):
    ex = normal_exchange
    ex.add(Message(role="user", content=[Text(text="User message 1")]))
    ex.reply()

    ex.pop_first_message()

    ex.add(Message(role="user", content=[Text(text="User message 2")]))
    ex.reply()

    assert len(ex.messages) == 3
    assert len(ex.checkpoint_data.checkpoints) == 3
    assert ex.checkpoint_data.message_index_offset == 1
    # (start, end)
    # (1, 1), (2, 2), (3, 3)
    # actual_index_in_messages_arr = any checkpoint index - offset
    assert no_overlapping_checkpoints(ex)
    for i in range(3):
        assert ex.checkpoint_data.checkpoints[i].start_index == i + 1
        assert ex.checkpoint_data.checkpoints[i].end_index == i + 1

    ex.pop_last_message()

    assert len(ex.messages) == 2
    assert len(ex.checkpoint_data.checkpoints) == 2
    assert ex.checkpoint_data.message_index_offset == 1
    assert no_overlapping_checkpoints(ex)
    for i in range(2):
        assert ex.checkpoint_data.checkpoints[i].start_index == i + 1
        assert ex.checkpoint_data.checkpoints[i].end_index == i + 1

    ex.add(Message(role="assistant", content=[Text(text="Assistant message")]))
    ex.add(Message(role="user", content=[Text(text="User message 3")]))
    ex.reply()

    assert len(ex.messages) == 5
    assert len(ex.checkpoint_data.checkpoints) == 4
    assert ex.checkpoint_data.message_index_offset == 1
    assert no_overlapping_checkpoints(ex)
    assert checkpoint_to_index_pairs(ex.checkpoint_data.checkpoints) == [(1, 1), (2, 2), (3, 4), (5, 5)]

    ex.pop_last_checkpoint()

    assert len(ex.messages) == 4
    assert len(ex.checkpoint_data.checkpoints) == 3
    assert ex.checkpoint_data.message_index_offset == 1
    assert no_overlapping_checkpoints(ex)
    assert checkpoint_to_index_pairs(ex.checkpoint_data.checkpoints) == [(1, 1), (2, 2), (3, 4)]

    ex.pop_first_message()

    assert len(ex.messages) == 3
    assert len(ex.checkpoint_data.checkpoints) == 2
    assert ex.checkpoint_data.message_index_offset == 2
    assert no_overlapping_checkpoints(ex)
    assert checkpoint_to_index_pairs(ex.checkpoint_data.checkpoints) == [(2, 2), (3, 4)]

    ex.pop_last_message()

    assert len(ex.messages) == 2
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert ex.checkpoint_data.message_index_offset == 2
    assert no_overlapping_checkpoints(ex)
    assert checkpoint_to_index_pairs(ex.checkpoint_data.checkpoints) == [(2, 2)]

    ex.pop_last_message()
    assert len(ex.messages) == 1
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert ex.checkpoint_data.message_index_offset == 2
    assert no_overlapping_checkpoints(ex)
    assert checkpoint_to_index_pairs(ex.checkpoint_data.checkpoints) == [(2, 2)]

    ex.add(Message(role="assistant", content=[Text(text="Assistant message")]))
    ex.add(Message(role="user", content=[Text(text="User message 5")]))
    ex.pop_last_checkpoint()

    assert len(ex.messages) == 0
    assert len(ex.checkpoint_data.checkpoints) == 0

    ex.add(Message(role="user", content=[Text(text="User message 6")]))
    ex.reply()

    assert len(ex.messages) == 2
    assert len(ex.checkpoint_data.checkpoints) == 2
    assert ex.checkpoint_data.message_index_offset == 2
    assert no_overlapping_checkpoints(ex)
    assert checkpoint_to_index_pairs(ex.checkpoint_data.checkpoints) == [(2, 2), (3, 3)]

    ex.pop_last_message()

    assert len(ex.messages) == 1
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert ex.checkpoint_data.message_index_offset == 2
    assert no_overlapping_checkpoints(ex)
    assert checkpoint_to_index_pairs(ex.checkpoint_data.checkpoints) == [(2, 2)]

    ex.pop_first_message()

    assert len(ex.messages) == 0
    assert len(ex.checkpoint_data.checkpoints) == 0
    assert ex.checkpoint_data.message_index_offset == 0

    ex.add(Message(role="user", content=[Text(text="User message 7")]))
    ex.pop_last_message()

    assert len(ex.messages) == 0
    assert len(ex.checkpoint_data.checkpoints) == 0
    assert ex.checkpoint_data.message_index_offset == 0


def test_pop_last_message_when_no_checkpoints_but_messages_present(normal_exchange):
    ex = normal_exchange
    ex.add(Message(role="user", content=[Text(text="User message")]))

    ex.pop_last_message()

    assert len(ex.messages) == 0
    assert len(ex.checkpoint_data.checkpoints) == 0
    assert ex.checkpoint_data.message_index_offset == 0


def test_pop_first_message_when_no_checkpoints_but_message_present(normal_exchange):
    ex = normal_exchange
    ex.add(Message(role="user", content=[Text(text="User message")]))

    with pytest.raises(ValueError) as e:
        ex.pop_first_message()

    assert str(e.value) == "There must be at least one checkpoint to pop the first message"


def test_pop_first_checkpoint_size_n(resumed_exchange):
    ex = resumed_exchange
    ex.pop_last_message()  # needed because the last message is an assistant message
    ex.reply()

    ex.pop_first_checkpoint()
    assert ex.checkpoint_data.message_index_offset == 5
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert len(ex.messages) == 1

    ex.pop_first_checkpoint()
    assert ex.checkpoint_data.message_index_offset == 0
    assert len(ex.checkpoint_data.checkpoints) == 0
    assert len(ex.messages) == 0


def test_pop_first_checkpoint_size_1(normal_exchange):
    ex = normal_exchange
    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()

    ex.pop_first_checkpoint()
    assert ex.checkpoint_data.message_index_offset == 1
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert len(ex.messages) == 1

    ex.pop_first_checkpoint()
    assert ex.checkpoint_data.message_index_offset == 0
    assert len(ex.checkpoint_data.checkpoints) == 0
    assert len(ex.messages) == 0


def test_pop_first_checkpoint_no_checkpoints(normal_exchange):
    ex = normal_exchange

    with pytest.raises(ValueError) as e:
        ex.pop_first_checkpoint()

    assert str(e.value) == "There are no checkpoints to pop"


def test_prepend_checkpointed_message_empty_exchange(normal_exchange):
    ex = normal_exchange
    ex.prepend_checkpointed_message(Message(role="assistant", content=[Text(text="Assistant message")]), 10)

    assert ex.checkpoint_data.message_index_offset == 0
    assert len(ex.checkpoint_data.checkpoints) == 1
    assert ex.checkpoint_data.checkpoints[0].start_index == 0
    assert ex.checkpoint_data.checkpoints[0].end_index == 0

    ex.add(Message(role="user", content=[Text(text="User message")]))
    ex.reply()

    assert ex.checkpoint_data.message_index_offset == 0
    assert len(ex.checkpoint_data.checkpoints) == 3
    assert len(ex.messages) == 3
    assert no_overlapping_checkpoints(ex)

    ex.pop_first_checkpoint()

    assert ex.checkpoint_data.message_index_offset == 1
    assert len(ex.checkpoint_data.checkpoints) == 2
    assert len(ex.messages) == 2
    assert no_overlapping_checkpoints(ex)

    ex.prepend_checkpointed_message(Message(role="assistant", content=[Text(text="Assistant message")]), 10)
    assert ex.checkpoint_data.message_index_offset == 0
    assert len(ex.checkpoint_data.checkpoints) == 3
    assert len(ex.messages) == 3
    assert no_overlapping_checkpoints(ex)


def test_generate_successful_response_on_first_try(normal_exchange):
    ex = normal_exchange
    ex.add(Message(role="user", content=[Text("Hello")]))
    ex.generate()


def test_rewind_in_normal_exchange(normal_exchange):
    ex = normal_exchange
    ex.rewind()

    assert len(ex.messages) == 0
    assert len(ex.checkpoint_data.checkpoints) == 0

    ex.add(Message(role="user", content=[Text("Hello")]))
    ex.generate()
    ex.add(Message(role="user", content=[Text("Hello")]))

    # testing if it works with a user text message at the end
    ex.rewind()

    assert len(ex.messages) == 2
    assert len(ex.checkpoint_data.checkpoints) == 2

    ex.add(Message(role="user", content=[Text("Hello")]))
    ex.generate()

    # testing if it works with a non-user text message at the end
    ex.rewind()

    assert len(ex.messages) == 2
    assert len(ex.checkpoint_data.checkpoints) == 2


def test_rewind_with_tool_usage():
    # simulating a real exchange with tool usage
    ex = Exchange(
        provider=MockProvider(
            sequence=[
                Message.assistant("Hello!"),
                Message(
                    role="assistant",
                    content=[ToolUse(id="1", name="dummy_tool", parameters={})],
                ),
                Message(
                    role="assistant",
                    content=[ToolUse(id="2", name="dummy_tool", parameters={})],
                ),
                Message.assistant("Done!"),
            ],
            usage_dicts=[
                {"usage": {"input_tokens": 12, "output_tokens": 23}},
                {"usage": {"input_tokens": 27, "output_tokens": 44}},
                {"usage": {"input_tokens": 50, "output_tokens": 56}},
                {"usage": {"input_tokens": 60, "output_tokens": 76}},
            ],
        ),
        model="gpt-4o-2024-05-13",
        system="You are a helpful assistant.",
        tools=[Tool.from_function(dummy_tool)],
        moderator=PassiveModerator(),
    )
    ex.add(Message(role="user", content=[Text(text="test")]))
    ex.reply()
    ex.add(Message(role="user", content=[Text(text="kick it off!")]))
    ex.reply()

    # removing the last message to simulate not getting a response
    ex.pop_last_message()

    # calling rewind to last user message
    ex.rewind()

    assert len(ex.messages) == 2
    assert len(ex.checkpoint_data.checkpoints) == 2
    assert no_overlapping_checkpoints(ex)
    assert ex.messages[0].content[0].text == "test"
    assert type(ex.messages[1].content[0]) is Text
    assert ex.messages[1].role == "assistant"



import pytest
from exchange import Exchange, Message
from exchange.content import ToolResult, ToolUse
from exchange.moderators.passive import PassiveModerator
from exchange.moderators.summarizer import ContextSummarizer
from exchange.providers import Usage


class MockProvider:
    def complete(self, model, system, messages, tools):
        assistant_message_text = "Summarized content here."
        output_tokens = len(assistant_message_text)
        total_input_tokens = sum(len(msg.text) for msg in messages)
        if not messages or messages[-1].role == "assistant":
            message = Message.user(assistant_message_text)
        else:
            message = Message.assistant(assistant_message_text)
        total_tokens = total_input_tokens + output_tokens
        usage = Usage(
            input_tokens=total_input_tokens,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )
        return message, usage


@pytest.fixture
def exchange_instance():
    ex = Exchange(
        provider=MockProvider(),
        model="test-model",
        system="test-system",
        messages=[
            Message.user("Hi, can you help me with my homework?"),
            Message.assistant("Of course! What do you need help with?"),
            Message.user("I need help with math problems."),
            Message.assistant("Sure, I can help with that. Let's get started."),
            Message.user("Can you also help with my science homework?"),
            Message.assistant("Yes, I can help with science too."),
            Message.user("That's great! How about history?"),
            Message.assistant("Of course, I can help with history as well."),
            Message.user("Thanks! You're very helpful."),
            Message.assistant("You're welcome! I'm here to help."),
        ],
        moderator=PassiveModerator(),
    )
    return ex


@pytest.fixture
def summarizer_instance():
    return ContextSummarizer(max_tokens=300)


def test_context_summarizer_rewrite(exchange_instance: Exchange, summarizer_instance: ContextSummarizer):
    # Pre-checks
    assert len(exchange_instance.messages) == 10

    exchange_instance.generate()

    # the exchange instance has a PassiveModerator so the messages are not truncated nor summarized
    assert len(exchange_instance.messages) == 11
    assert len(exchange_instance.checkpoint_data.checkpoints) == 2

    # we now tell the summarizer to summarize the exchange
    summarizer_instance.rewrite(exchange_instance)

    assert exchange_instance.checkpoint_data.total_token_count <= 200
    assert len(exchange_instance.messages) == 2

    # Assert that summarized content is the first message
    first_message = exchange_instance.messages[0]
    assert first_message.role == "user" or first_message.role == "assistant"
    assert any("summarized" in content.text.lower() for content in first_message.content)

    # Ensure roles alternate in the output
    for i in range(1, len(exchange_instance.messages)):
        assert (
            exchange_instance.messages[i - 1].role != exchange_instance.messages[i].role
        ), "Messages must alternate between user and assistant"


MESSAGE_SEQUENCE = [
    Message.user("Hi, can you help me with my homework?"),
    Message.assistant("Of course! What do you need help with?"),
    Message.user("I need help with math problems."),
    Message.assistant("Sure, I can help with that. Let's get started."),
    Message.user("What is 2 + 2, 3*3, 9/5, 2*20, 14/2?"),
    Message(
        role="assistant",
        content=[ToolUse(id="1", name="add", parameters={"a": 2, "b": 2})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="1", output="4")]),
    Message(
        role="assistant",
        content=[ToolUse(id="2", name="multiply", parameters={"a": 3, "b": 3})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="2", output="9")]),
    Message(
        role="assistant",
        content=[ToolUse(id="3", name="divide", parameters={"a": 9, "b": 5})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="3", output="1.8")]),
    Message(
        role="assistant",
        content=[ToolUse(id="4", name="multiply", parameters={"a": 2, "b": 20})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="4", output="40")]),
    Message(
        role="assistant",
        content=[ToolUse(id="5", name="divide", parameters={"a": 14, "b": 2})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="5", output="7")]),
    Message.assistant("I'm done calculating the answers to your math questions."),
    Message.user("Can you also help with my science homework?"),
    Message.assistant("Yes, I can help with science too."),
    Message.user("What is the speed of light? The frequency of a photon? The mass of an electron?"),
    Message(
        role="assistant",
        content=[ToolUse(id="6", name="speed_of_light", parameters={})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="6", output="299,792,458 m/s")]),
    Message(
        role="assistant",
        content=[ToolUse(id="7", name="photon_frequency", parameters={})],
    ),
    Message(role="user", content=[ToolResult(tool_use_id="7", output="2.418 x 10^14 Hz")]),
    Message(role="assistant", content=[ToolUse(id="8", name="electron_mass", parameters={})]),
    Message(
        role="user",
        content=[ToolResult(tool_use_id="8", output="9.10938356 x 10^-31 kg")],
    ),
    Message.assistant("I'm done calculating the answers to your science questions."),
    Message.user("That's great! How about history?"),
    Message.assistant("Of course, I can help with history as well."),
    Message.user("Thanks! You're very helpful."),
    Message.assistant("You're welcome! I'm here to help."),
]


class AnotherMockProvider:
    def __init__(self):
        self.sequence = MESSAGE_SEQUENCE
        self.current_index = 1
        self.summarize_next = False
        self.summarized_count = 0

    def complete(self, model, system, messages, tools):
        system_prompt_tokens = 100
        input_token_count = system_prompt_tokens

        message = self.sequence[self.current_index]
        if self.summarize_next:
            text = "Summary message here"
            self.summarize_next = False
            self.summarized_count += 1
            return Message.assistant(text=text), Usage(
                # in this case, input tokens can really be whatever
                input_tokens=40,
                output_tokens=len(text) * 2,
                total_tokens=40 + len(text) * 2,
            )

        if len(messages) > 0 and type(messages[0].content[0]) is ToolResult:
            raise ValueError("ToolResult should not be the first message")

        if len(messages) == 1 and messages[0].text == "a":
            # adding a +1 for the "a"
            return Message.assistant("Getting system prompt size"), Usage(
                input_tokens=80 + 1, output_tokens=20, total_tokens=system_prompt_tokens + 1
            )

        for i in range(len(messages)):
            if type(messages[i].content[0]) in (ToolResult, ToolUse):
                input_token_count += 10
            else:
                input_token_count += len(messages[i].text) * 2

        if type(message.content[0]) in (ToolResult, ToolUse):
            output_tokens = 10
        else:
            output_tokens = len(message.text) * 2

        total_tokens = input_token_count + output_tokens
        if total_tokens > 300:
            self.summarize_next = True
        usage = Usage(
            input_tokens=input_token_count,
            output_tokens=output_tokens,
            total_tokens=total_tokens,
        )
        self.current_index += 2
        return message, usage


@pytest.fixture
def conversation_exchange_instance():
    ex = Exchange(
        provider=AnotherMockProvider(),
        model="test-model",
        system="test-system",
        moderator=ContextSummarizer(max_tokens=300),
        # TODO: make it work with an offset so we don't have to send off requests basically
        # at every generate step
    )
    return ex


def test_summarizer_generic_conversation(conversation_exchange_instance: Exchange):
    i = 0
    while i < len(MESSAGE_SEQUENCE):
        next_message = MESSAGE_SEQUENCE[i]
        conversation_exchange_instance.add(next_message)
        message = conversation_exchange_instance.generate()
        if message.text != "Summary message here":
            i += 2
    checkpoints = conversation_exchange_instance.checkpoint_data.checkpoints
    assert conversation_exchange_instance.checkpoint_data.total_token_count == 570
    assert len(checkpoints) == 10
    assert len(conversation_exchange_instance.messages) == 10
    assert checkpoints[0].start_index == 20
    assert checkpoints[0].end_index == 20
    assert checkpoints[-1].start_index == 29
    assert checkpoints[-1].end_index == 29
    assert conversation_exchange_instance.checkpoint_data.message_index_offset == 20
    assert conversation_exchange_instance.provider.summarized_count == 12
    assert conversation_exchange_instance.moderator.system_prompt_token_count == 100



import os

import pytest

from exchange import Text, ToolUse
from exchange.providers.openai import OpenAiProvider
from .conftest import complete, vision, tools

OPENAI_MODEL = os.getenv("OPENAI_MODEL", "gpt-4o-mini")


@pytest.mark.vcr()
def test_openai_complete(default_openai_env):
    reply_message, reply_usage = complete(OpenAiProvider, OPENAI_MODEL)

    assert reply_message.content == [Text(text="Hello! How can I assist you today?")]
    assert reply_usage.total_tokens == 27


@pytest.mark.integration
def test_openai_complete_integration():
    reply = complete(OpenAiProvider, OPENAI_MODEL)

    assert reply[0].content is not None
    print("Completion content from OpenAI:", reply[0].content)


@pytest.mark.vcr()
def test_openai_tools(default_openai_env):
    reply_message, reply_usage = tools(OpenAiProvider, OPENAI_MODEL)

    tool_use = reply_message.content[0]
    assert isinstance(tool_use, ToolUse)
    assert tool_use.id == "call_xXYlw4A7Ud1qtCopuK5gEJrP"
    assert tool_use.name == "read_file"
    assert tool_use.parameters == {"filename": "test.txt"}
    assert reply_usage.total_tokens == 122


@pytest.mark.integration
def test_openai_tools_integration():
    reply = tools(OpenAiProvider, OPENAI_MODEL)

    tool_use = reply[0].content[0]
    assert isinstance(tool_use, ToolUse)
    assert tool_use.id is not None
    assert tool_use.name == "read_file"
    assert tool_use.parameters == {"filename": "test.txt"}


@pytest.mark.vcr()
def test_openai_vision(default_openai_env):
    reply_message, reply_usage = vision(OpenAiProvider, OPENAI_MODEL)

    assert reply_message.content == [Text(text='The first entry in the menu says "Ask Goose."')]
    assert reply_usage.total_tokens == 14241


@pytest.mark.integration
def test_openai_vision_integration():
    reply = vision(OpenAiProvider, OPENAI_MODEL)

    assert "ask goose" in reply[0].text.lower()



from collections import defaultdict
from typing import Dict

from exchange.providers.base import Usage


class _TokenUsageCollector:
    def __init__(self) -> None:
        self.usage_data = []

    def collect(self, model: str, usage: Usage) -> None:
        self.usage_data.append((model, usage))

    def get_token_usage_group_by_model(self) -> Dict[str, Usage]:
        usage_group_by_model = defaultdict(lambda: Usage(0, 0, 0))
        for model, usage in self.usage_data:
            usage_by_model = usage_group_by_model[model]
            if usage is not None and usage.input_tokens is not None:
                usage_by_model.input_tokens += usage.input_tokens
            if usage is not None and usage.output_tokens is not None:
                usage_by_model.output_tokens += usage.output_tokens
            if usage is not None and usage.total_tokens is not None:
                usage_by_model.total_tokens += usage.total_tokens
        return usage_group_by_model


_token_usage_collector = _TokenUsageCollector()
